#!/bin/bash
# Copyright 2013 - 2015 Michal Žejdl
# standalone non parallel (mild) dual stack reverse resolver
# called from: start.sh
# input: arp.ip, nd.ip6 (sql)
# output: rr4, rr6 (sql tables)
# dependence: mysql dig gawk

confdir=$(dirname $0)
config=$confdir/defaults.sh
test -f $config || exit
. $config

# smaller limit leads to heavier mysql load
limit=100
# save IFS value (one time)
oldifs=$IFS

# a new file descriptor to prevent pipe from closing
exec 3> $mysqlwpipe

do_query () {
	test -z "$1" && return 1
	result=($(mysql -e "$1" $select_opts)) || return 1
	#echo "$result"
}

# is in arp and not in rr4
get_new4 () {
	do_query "select distinct arp.ip from arp left join rr4 on arp.ip = rr4.address where rr4.address is null order by arp.time desc limit $limit" || return 1
}

# is in nd and not in rr6
get_new6 () {
	# select only global (2000::/3); how to avoid RFC 4941 addresses?
	do_query "select distinct nd.ip6 from nd left join rr6 on nd.ip6 = rr6.address where rr6.address is null and (nd.ip6 like '2%' or nd.ip6 like '3%') order by nd.time desc limit $limit" || return 1
}

# TTL expired
get_old4 () {
	do_query "select distinct address from rr4 where ttl is not null and unix_timestamp() > time + ttl order by time limit $limit" || return 1
}

get_old6 () {
	do_query "select distinct address from rr6 where ttl is not null and unix_timestamp() > time + ttl order by time limit $limit" || return 1
}

# the oldest noname in rr4 with a new arp entry 
get_noname_live4 () {
	do_query "select distinct address from rr4 join arp on rr4.address = arp.ip where rr4.name is null and rr4.time < arp.time order by rr4.time limit $limit" || return 1
}

# the oldest noname in rr6 with a new nd entry 
get_noname_live6 () {
	do_query "select distinct address from rr6 join nd on rr6.address = nd.ip6 where rr6.name is null and rr6.time < nd.time order by rr6.time limit $limit" || return 1
}

# the oldest noname in rr4 without a new arp entry
# take only those which have been attempted to resolve (rr4.time)
#	closer to last live (arp.time) than now (unix_timestamp())
#	to reduce unnecessary traffic to resolver
get_noname_dead_diminished4 () {
	do_query "select distinct address from rr4 join arp on rr4.address = arp.ip where rr4.name is null and rr4.time >= arp.time and rr4.time - arp.time < unix_timestamp() - rr4.time order by rr4.time limit $limit" || return 1
}

# the oldest noname in rr6 without a new nd entry
# take only those which have been attempted to resolve (rr6.time)
#	closer to last live (nd.time) than now (unix_timestamp())
#	to reduce unnecessary traffic to resolver
get_noname_dead_diminished6 () {
	do_query "select distinct address from rr6 join nd on rr6.address = nd.ip6 where rr6.name is null and rr6.time >= nd.time and rr6.time - nd.time < unix_timestamp() - rr6.time order by rr6.time limit $limit" || return 1
}

# input: IP address, version (4 or 6); output: insert/update rrversion table
update () {
	updatetime=$(date +%s);
	#answer="$($digptr $1)" || return 1
	# dig -x can return more lines (PTR records)
	# -> array is mandatory and only \n is a delimiter
	# switch IFS only here, it changes $"value with spaces" behaviour
	#answer=($($digptr $1)) || answer=()
	IFS=$'\n'
	# empty answer on dig error
	answer=($(dig +noall +answer -x $1)) || answer=()
	IFS=$oldifs
	#if test -z "$answer"; then
	if test ${#answer[@]} -eq 0; then
		# dig returned empty output or error
		# echo should be fflushed automaticaly according to http://stackoverflow.com/questions/9926616/is-echo-atomic-when-writing-single-lines
		echo "insert rr$2 (address, name, ttl, time) values ('$1', null, null, unix_timestamp()) on duplicate key update name = null, ttl = null, time = unix_timestamp();" >&3
	else
		for ((ai = 0; ai < ${#answer[@]}; ai++)); do
			#echo "$answer" | gawk -v address=$1 -v version=$2 '{print "insert rr"version" (address, name, ttl, time) values (\47"address"\47, \47"substr($5, 1, length($5) - 1)"\47, "$2", unix_timestamp()) on duplicate key update name = \47"substr($5, 1, length($5) - 1)"\47, ttl = "$2", time = unix_timestamp();"; fflush()}' >&3
			echo "${answer[$ai]}" | gawk -v address=$1 -v version=$2 '{print "insert rr"version" (address, name, ttl, time) values (\47"address"\47, \47"substr($5, 1, length($5) - 1)"\47, "$2", unix_timestamp()) on duplicate key update name = \47"substr($5, 1, length($5) - 1)"\47, ttl = "$2", time = unix_timestamp();"; fflush()}' >&3
		done
	fi
	# delete the old rows manually because an address is not unique
	echo "delete from rr$2 where address = '$1' and time < $updatetime;" >&3
}

while true; do
	for method in new old noname_live noname_dead_diminished; do
		for version in 4 6; do
			if get_$method$version; then
				for ((ri = 0; ri < ${#result[@]}; ri++)); do
  					# stop this golem if "šém" is removed
					test -e $rundir/$(basename $0 .sh).sem || exit
					update "${result[$ri]}" "$version"
					sleep 1
				done
			fi
		done
	done
	#sleep 1;
done
