#!/bin/bash
# Copyright 2015 Michal Žejdl
# try to distinguish SNMP agents using RFC 1213 system MIB
#	snmpEngineID is SNMPv3 only and not so smart (MAC address or algorithm)
# called by: start.sh (as a golem)
# input: snmp4, snmp6, agent (sql tables), SNMP agents
# output: snmp4, snmp6, agent (sql tables)
# dependence: dirname, mysql, snmpget, date

confdir=$(dirname $0)/..
config=$confdir/defaults.sh
test -f $config || exit
. $config

# smaller limit leads to heavier mysql load
limit=100

# it took 2 secs for Westermo 225 to answer snmpgetnext
# no CPU high on switch, strange...
#
# cite:
# "Sad to say the 2 to 2.3 second delay is well known and it will not be fixed.
# ...
# This could have been done in a better way had the issue been known
# when the hardware design for Lynx was done but now we have to make do
# with what we got."
#
# so here is an workaround
snmpgetnexttimeout=3

# a new file descriptor to prevent pipe from closing
# uncomment after debugging
exec 3> $mysqlwpipe

do_query () {
	test -z "$1" && return 1
	result=($(mysql -e "$1" $select_opts)) || return 1
}

# has SNMP, but no associated agent
get_snmpaddress4 () {
	do_query "select address from snmp4 where status = 0 and agent is null order by time limit $limit" || return 1
}

get_snmpaddress6 () {
	do_query "select address from snmp6 where status = 0 and agent is null order by time limit $limit" || return 1
}

get_oldest_agent_with_oldest_addresses ()
{
	# > $snmpinterval to avoid spam-checking if there are a few agents
	# status = 0 only; addresses.sh does the checking job
	# limit 1 is for simplicity (to handle bash 1D array)
	#	assuming, that there will be much less agents than addresses,
	# 	so db load wouldn't be so hard
	#	let me know, if I'm wrong
	#do_query "select id, snmp4.address, snmp4.get, snmp6.address, snmp6.get from snmpagent left join snmp4 on id=snmp4.agent left join snmp6 on id=snmp6.agent where unix_timestamp() - snmpagent.time > 300 and (snmp4.address is not null and snmp4.status = 0 or snmp6.address is not null and snmp6.status = 0) order by snmpagent.time, snmp4.get, snmp6.get limit 1" || return 1
	# this query is wrong - it doesn't include cases when 4 address is OK
	# (status 0) and 6 address failed (status > 0) and vice versa
	#do_query "select id, snmp4.address, snmp4.get, snmp6.address, snmp6.get from snmpagent left join snmp4 on id=snmp4.agent left join snmp6 on id=snmp6.agent where unix_timestamp() - snmpagent.time > $snmpinterval and (snmp4.status = 0 and snmp6.status is null or snmp4.status is null and snmp6.status = 0 or snmp4.status = 0 and snmp6.status = 0) order by snmpagent.time, snmp4.get, snmp6.get limit 1" || return 1
	# select oldest agent with valid address and pick 4 or 6 later
	do_query "select id, snmp4.address, snmp4.status, snmp4.get, snmp6.address, snmp6.status, snmp6.get from snmpagent left join snmp4 on id=snmp4.agent left join snmp6 on id=snmp6.agent where unix_timestamp() - snmpagent.time > $snmpinterval and (snmp4.status = 0 or snmp6.status = 0) order by snmpagent.time, snmp4.get, snmp6.get limit 1" || return 1
}

# not used - the function above does the job
get_oldest_agent_with_oldest_address4 ()
{
	do_query "select id, address from snmpagent join snmp4 on id = agent where snmpagent.time < unix_timestamp() - 300 and status =  0 order by snmpagent.time, snmp4.get limit 1" || return 1
}

get_oldest_agent_with_oldest_address6 ()
{
	do_query "select id, address from snmpagent join snmp6 on id = agent where snmpagent.time < unix_timestamp() - 300 and status =  0 order by snmpagent.time, snmp6.get limit 1" || return 1
}

parse_eight_oids ()
{
	for ((oid_row = 0; oid_row < ${#eight_oids[@]}; oid_row++)); do
		# this approach takes only the first line; rewrite later
		case ${eight_oids[$oid_row]::18} in
		.1.3.6.1.2.1.1.1.0)
			# bash 4.1.2 substring removal slowly eats the memory
			# this script consumes several hundreds MB (in RSS)
			# and exits on "fork: Cannot allocate memory"
			# so for quick fix I use subshell (echo "...")
			# to exchange memory for cpu cycles
			#
			# without MIB (-m '') there are quotes around strings
			#sysDescr=${eight_oids[oid_row]#.1.3.6.1.2.1.1.1.0 \"}
			sysDescr=$(echo "${eight_oids[oid_row]#.1.3.6.1.2.1.1.1.0 \"}")
			#sysDescr=${sysDescr%\"}
			sysDescr=$(echo "${sysDescr%\"}")
			;;
		.1.3.6.1.2.1.1.2.0)
			sysObjectID=$(echo "${eight_oids[oid_row]#.1.3.6.1.2.1.1.2.0 }")
			;;
		.1.3.6.1.2.1.1.3.0)
			sysUpTime=$(echo "${eight_oids[oid_row]#.1.3.6.1.2.1.1.3.0 }")
			;;
		.1.3.6.1.2.1.1.4.0)
			sysContact=$(echo "${eight_oids[oid_row]#.1.3.6.1.2.1.1.4.0 \"}")
			sysContact=$(echo "${sysContact%\"}")
			;;
		.1.3.6.1.2.1.1.5.0)
			sysName=$(echo "${eight_oids[oid_row]#.1.3.6.1.2.1.1.5.0 \"}")
			sysName=$(echo "${sysName%\"}")
			;;
		.1.3.6.1.2.1.1.6.0)
			sysLocation=$(echo "${eight_oids[oid_row]#.1.3.6.1.2.1.1.6.0 \"}")
			sysLocation=$(echo "${sysLocation%\"}")
			;;
		.1.3.6.1.2.1.1.7.0)
			sysServices=$(echo "${eight_oids[oid_row]#.1.3.6.1.2.1.1.7.0 }")
			;;
		.1.3.6.1.2.1.2.1.0)
			ifNumber=$(echo "${eight_oids[oid_row]#.1.3.6.1.2.1.2.1.0 }")
			;;
		esac
	done
}

snmpget_one_oid ()
{
	# -r 0, because there was at least one packet short time before
	# -Oa to get rid of Hex-STRING triggered by some chars
	# 2>/dev/null - here will be at least one "Error in packet" per agent
	# so don't bother with spamming stderr
	one_oid=$(snmpget -m '' -v 1 -r 0 -c $snmprcommunity -Oanqtv $snmpaddress $1 2>/dev/null)
}

snmpget_string ()
{
	snmpget_one_oid $1 || return
	# avoid a memleak with the subshell; see parse_eight_oids ()
	# remove the opening quotes
	one_oid=$(echo "${one_oid#\"}")
	# remove the ending quotes
	one_oid=$(echo "${one_oid%\"}")
	# remove possible apostrophes to avoid mysql injection
	#one_oid=${one_oid//\'}
}

try_to_get_eight_one_by_one ()
{
	local one_oid
	snmpget_string .1.3.6.1.2.1.1.1.0 && sysDescr=$one_oid
	snmpget_one_oid .1.3.6.1.2.1.1.2.0 && sysObjectID=$one_oid
	snmpget_string .1.3.6.1.2.1.1.4.0 && sysContact=$one_oid
	snmpget_string .1.3.6.1.2.1.1.5.0 && sysName=$one_oid
	snmpget_string .1.3.6.1.2.1.1.6.0 && sysLocation=$one_oid
	snmpget_one_oid .1.3.6.1.2.1.1.7.0 && sysServices=$one_oid
	snmpget_one_oid .1.3.6.1.2.1.2.1.0 && ifNumber=$one_oid
	# last one will be sysUpTime to get close to date timestamp below
	snmpget_one_oid .1.3.6.1.2.1.1.3.0 && sysUpTime=$one_oid
	# sysUpTime is most needed - to distinguish when all values are default
	# test != leads to return the ok status on the end of this function
	#test $sysUpTime != "NULL" || return 2
}

replace_specchars ()
{
	# memleak appears in substring replacement as well
	# see parse_eight_oids ()
	# avoid mysql injection via string variables
	without_specchars=$(echo "${1//\'/ }")
	# do not cope with a newline
	without_specchars=$(echo "${without_specchars//$'\n'/ }")
	# replace carriage return as well
	without_specchars=$(echo "${without_specchars//$'\r'/ }")
}

is_natural_number ()
{
	test "$1" -ge 0 2>/dev/null
}

test_dot1dTpFdbEntry ()
{
	# test .1.3.6.1.2.1.17.4.3.1.2 (dot1dTpFdbPort) used in fdb.sh
	local result=$(snmpgetnext -m '' -v 1 -r 0 -t $snmpgetnexttimeout -c $snmprcommunity -Onq $snmpaddress .1.3.6.1.2.1.17.4.3.1.2 2>/dev/null) || return
	test "${result::23}" = .1.3.6.1.2.1.17.4.3.1.2
}

test_ifXEntry ()
{
	# test .1.3.6.1.2.1.31.1.1.1.2 (ifInMulticastPkts)
	# the first counter (non string) in RFC 1573, 2233 ifXTable/ifXEntry
	# I read somewhere (Cisco?), that -v 2c is mandatory for ifX*
	#	but net-snmp on CentOS 5.11 is happy with -v 1
	local result=$(snmpgetnext -m '' -v 1 -r 0 -t $snmpgetnexttimeout -c $snmprcommunity -Onq $snmpaddress .1.3.6.1.2.1.31.1.1.1.2 2>/dev/null) || return
	test "${result::23}" = .1.3.6.1.2.1.31.1.1.1.2
}

test_agent_capability ()
{
	dot1dTpFdbEntry="n"
	ifXEntry="n"
	test_dot1dTpFdbEntry && dot1dTpFdbEntry="y"
	test_ifXEntry && ifXEntry="y"
}

# input: IP address, version (4 or 6); output: $sys* and $ifNumber variables
read_snmp_agent ()
{
	if test $2 -eq 6; then
		snmpaddress=udp6:[$1]
	else
		snmpaddress=$1
	fi
	# unknown int variables
	sysUpTime=0
	sysServices=0
	ifNumber=0
	# unknown strings remain empty (the apostrophes do the job)
	unset sysDescr sysObjectID sysContact sysName sysLocation
	# hunting memleak
	unset eight_oids
	oldifs=$IFS
	# one line = one array element; and focus only on the first line
	IFS=$'\n'
	# one retry (-r 1) for the datagram discard sake
	# do not resend request if OID is missing (-Cf); handle it later
	# - there are buggy agents with more OIDs missing and strange behavior
	# -Oa to get rid of Hex-STRING triggered by some chars
	# read system 1 - 7 and ifNumber togethether
	# - all are mandatory, so craft one PDU for maximum efficiency
	eight_oids=($(snmpget -m '' -v 1 -r 1 -c $snmprcommunity -Cf -Oanqt $snmpaddress .1.3.6.1.2.1.1.1.0 .1.3.6.1.2.1.1.2.0 .1.3.6.1.2.1.1.3.0 .1.3.6.1.2.1.1.4.0 .1.3.6.1.2.1.1.5.0 .1.3.6.1.2.1.1.6.0 .1.3.6.1.2.1.1.7.0 .1.3.6.1.2.1.2.1.0))
	local snmpget_status=$?
	IFS=$oldifs
	case $snmpget_status in
	0)
		parse_eight_oids
		;;
	1)
		# Timeout: No Response from ...
		echo "$(date '+%F %X') snmpget $1 timeout" >&2
		;;
	2)
		# Error in packet
		# Reason: (noSuchName) There is no such variable name in this MIB.
		# Failed object: .1.3.6.1.2.1...
		echo "$(date '+%F %X') snmpget $1 error in PDU" >&2
		# if try_ succeeds, change the status to ok
		try_to_get_eight_one_by_one && snmpget_status=0
		;;
	*)
		echo "$(date '+%F %X') unknown snmpget $1 exit status: $snmpget_status" >&2
		;;
	esac
	# to speed things up
	test $snmpget_status -ne 0 && return $snmpget_status
	# unix_timestamp() in query is sufficient
	#uptimestamp=$(date +%s)
	# sanity checking
	replace_specchars "$sysDescr" && sysDescr=$without_specchars
	replace_specchars "$sysObjectID" && sysObjectID=$without_specchars
	if ! is_natural_number "$sysUpTime"; then
		echo "$(date '+%F %X') $1 returned weird sysUpTime: $sysUpTime"
		sysUpTime=0
	fi
	replace_specchars "$sysContact" && sysContact=$without_specchars
	replace_specchars "$sysName" && sysName=$without_specchars
	replace_specchars "$sysLocation" && sysLocation=$without_specchars
	if ! is_natural_number "$sysServices"; then
		echo "$(date '+%F %X') $1 returned weird sysServices: $sysServices"
		sysServices=0
	fi
	# sysObjectID .1.3.6.1.4.1.21796.4.6 returned this ifNumber:
	#	No Such Object available on this agent at this OID
	#	(without any error)
	if ! is_natural_number "$ifNumber"; then
		echo "$(date '+%F %X') $1 returned weird ifNumber: $ifNumber"
		ifNumber=0
	fi
	test_agent_capability
	return $snmpget_status
}

# input: IP address, version (4 or 6), status; output: updated snmp[46] table
disable_address ()
{
	#echo "$(date '+%F %X') disable IPv$2 address $1 with status $3"
	echo "update snmp$2 set time = unix_timestamp(), status = $3 where address = '$1';" >&3
}

# in case the uniqueness checking in find_agent() is changed
delete_agent ()
{
	echo "\
delete from snmp4 where agent = $1; \
delete from snmp6 where agent = $1; \
delete from snmpagent where id = $1; \
delete from mibapc where agent = $1; \
delete from mibifentry where agent = $1; \
delete from mibifxentry where agent = $1; \
	" >&3
}

find_agent ()
{
	# handle null numbers
	#if test $sysServices = "NULL"; then
	#	cond_sysServices="sysServices is null"
	#else
	#	cond_sysServices="sysServices = $sysServices"
	#fi
	#if test $ifNumber = "NULL"; then
	#	cond_ifNumber="ifNumber is null"
	#else
	#	cond_ifNumber="ifNumber = $ifNumber"
	#fi
	# ts2 - ts1 ~ (ut2 - ut1) div 100
	# 	~ is equality with 10 second tolerance (load, RTT, clock skew)
	#		10 is way to small for HP printers which skew is great
	#		trying 100 (more defaults may be bundled)
	# 	ut2 can be smaller than ut1 (reboot, 32b counter overflow)
	# 3 - agent not found due to mysql error
	#agentids=($(mysql -e "select id from snmpagent where abs($uptimestamp - time - cast($sysUpTime - sysUpTime as signed) div 100) < 100 and sysDescr = '$sysDescr' and sysObjectID = '$sysObjectID' and sysContact = '$sysContact' and sysName = '$sysName' and sysLocation = '$sysLocation' and $cond_sysServices and $cond_ifNumber" $select_opts)) || return 3
	#agentids=($(mysql -e "select id from snmpagent where $sysUpTime >= sysUpTime and cast($sysUpTime - sysUpTime as signed) div 100 <= $uptimestamp - time + 10 and sysDescr = '$sysDescr' and sysObjectID = '$sysObjectID' and sysContact = '$sysContact' and sysName = '$sysName' and sysLocation = '$sysLocation' and $cond_sysServices and $cond_ifNumber" $select_opts)) || return 3
	# query on unique index; defaults can't be distinguished (lazy admins)
	# give up building uniqueness on sysUpTime
	#	- skew on some (HP printer) agents is secs/mins (sleeping?)
	#agentids=($(mysql -e "select id from snmpagent where sysDescr = '$sysDescr' and sysObjectID = '$sysObjectID' and sysContact = '$sysContact' and sysName = '$sysName' and sysLocation = '$sysLocation' and sysServices = $sysServices and ifNumber = $ifNumber" $select_opts)) || return 3
	# same with ifNumber - changes a lot on some agents (pppoe server)
	agentids=($(mysql -e "select id from snmpagent where sysDescr = '$sysDescr' and sysObjectID = '$sysObjectID' and sysContact = '$sysContact' and sysName = '$sysName' and sysLocation = '$sysLocation' and sysServices = $sysServices" $select_opts)) || return 3
	case ${#agentids[@]} in
	0)
		# no agent found
		#"$(date '+%F %X') find_agent: query returned no row"
		#echo mysql -e "select id from snmpagent where sysDescr = '$sysDescr' and sysObjectID = '$sysObjectID' and sysContact = '$sysContact' and sysName = '$sysName' and sysLocation = '$sysLocation' and sysServices = $sysServices and ifNumber = $ifNumber"
		;;
	1)
		agentid=${agentids[0]}
		;;
	*)
		# 2 - more agents found
		echo -n "$(date '+%F %X') find_agent(): more agents found:"
		for ((dupindex = 0; dupindex < ${#agentids[@]}; dupindex++)); do
			echo -n " ${agentids[dupindex]}"
			if test $dupindex -gt 0; then
				delete_agent ${agentids[dupindex]}
				echo -n " (deleted)"
			fi
		done
		echo " - check the algorithm for uniqueness"
		return 2
		;;
	esac
	return ${#agentids[@]}
}

insert_a_new_agent ()
{
	echo "insert snmpagent (time, sysDescr, sysObjectID, sysUpTime, sysContact, sysName, sysLocation, sysServices, ifNumber, dot1dTpFdbEntry, ifXEntry) values (unix_timestamp(), '$sysDescr', '$sysObjectID', $sysUpTime, '$sysContact', '$sysName', '$sysLocation', $sysServices, $ifNumber, '$dot1dTpFdbEntry', '$ifXEntry');" >&3
}

#insert_or_update_agent ()
#{
#	echo "insert snmpagent (time, sysDescr, sysObjectID, sysUpTime, sysContact, sysName, sysLocation, sysServices, ifNumber) values (unix_timestamp(), '$sysDescr', '$sysObjectID', $sysUpTime, '$sysContact', '$sysName', '$sysLocation', $sysServices, $ifNumber) on duplicate key update time = unix_timestamp(), sysUpTime = $sysUpTime;" >&3
#}

#link_address_to_agent ()
#{
#	echo "update snmp$2 set agent = $3, get = unix_timestamp() where address = '$1';" >&3
#}

#update_address_get ()
#{
#	echo "update snmp$2 set get = unix_timestamp() where address = '$1';" >&3
#}

update_agent_and_address ()
{
	#echo "$(date '+%F %X') agent $1 found for IPv$3 address $2"
	# update non unique values (see unique checking in find_agent())
	echo "update snmpagent set time = unix_timestamp(), sysUpTime = $sysUpTime, dot1dTpFdbEntry = '$dot1dTpFdbEntry', ifXEntry = '$ifXEntry' where id = $1;" >&3
	echo "update snmp$3 set agent = $1, get = unix_timestamp() where address = '$2';" >&3
}

# input address, version
unlink_address_from_agent ()
{
	echo "update snmp$2 set agent = null, get = unix_timestamp() where address = '$1';" >&3
}

while true; do
	for version in 4 6; do
		# hunting a memleak
		unset result;
		if get_snmpaddress$version; then
			for ((ri = 0; ri < ${#result[@]}; ri++)); do
  				# stop this golem if "šém" is removed
				test -e $rundir/snmp/$(basename $0 .sh).sem || exit
				read_snmp_agent "${result[$ri]}" $version
				status=$?
				if test $status -ne 0; then
					disable_address "${result[$ri]}" $version $status
					# no need for sleep here
					# snmpget did this (retry)
					continue
				fi
				#insert_or_update_agent
				find_agent
				case $? in
				0)
					# get dot1d, ifX
					#echo "$(date '+%F %X') a new agent found on address ${result[$ri]}"
					insert_a_new_agent
					# link to address in next while loop
					;;
				1)
					#link_address_to_agent "${result[$ri]}" $version $agentid
					update_agent_and_address $agentid "${result[$ri]}" $version
					;;
				2)
					echo "$(date '+%F %X') more matching agents found, check the condition in find_agent" >&2
					;;
				3)
					echo "$(date '+%F %X') mysql error" >&2
					;;
				esac
				# cleanup the find_agent() variables
				# memleak hunting
				unset agentid agentids
				sleep 1
			done
		fi
	done

  	# stop this golem if "šém" is removed
	test -e $rundir/snmp/$(basename $0 .sh).sem || exit

	# hunting a memleak
	unset used_address;
	unset result;
	get_oldest_agent_with_oldest_addresses
	# 7 values expected - see the query in function above
	# id, 4address, 4status, 4gettime, 6address, 6status, 6gettime
	if test ${#result[@]} -ne 7; then
		echo "$(date '+%F %X') get_oldest_agent_with_oldest_addresses() returned ${#result[@]} values, but 7 expected" >&2
		sleep 1
		continue
	fi
	# select oldest IP address from 4address and 6address
	if test ${result[2]} != "0"; then
		#echo "IPv6 only"
		used_address=${result[4]}
		version=6
	elif test ${result[5]} != "0"; then
		#echo "IPv4 only"
		used_address=${result[1]}
		version=4
	elif test ${result[3]} -lt ${result[6]}; then
		#echo "IPv4 is older"
		used_address=${result[1]}
		version=4
	else
		#echo "IPv6 is equal or older"
		used_address=${result[4]}
		version=6
	fi
	read_snmp_agent "$used_address" $version
	status=$?
	if test $status -ne 0; then
		disable_address "$used_address" $version $status
		# no need for sleep here
		# snmpget did this (retry)
		continue
	fi
	#insert_or_update_agent
	find_agent
	case $? in
	0)
		# get dot1d, ifX
		#echo "$(date '+%F %X') a new agent found on address $used_address"
		unlink_address_from_agent "$used_address" $version
		insert_a_new_agent
		# link to address in next while loop
		;;
	1)
		# same agent id found in db as was the former
		if test $agentid -eq ${result[0]}; then
			update_agent_and_address $agentid "$used_address" $version
			#update_address_get "$used_address" $version
		else
			unlink_address_from_agent "$used_address" $version
			# link to agent in the next loop
		fi
		;;
	2)
		echo "$(date '+%F %X') more matching agents found, check the condition in find_agent" >&2
		;;
	3)
		echo "$(date '+%F %X') mysql error" >&2
		;;
	esac
	sleep 1
done
