/*
 * the first algorithm, which may suffer by MAC flapping:
 * - when adding second (and next) address to the port, combine this address
 *   to pair with addresses added earlier and try to find bridge, which has
 *   these two addresses on distinct ports; this bridge is laying behind
 *   the port(s) of bridge(s), where these two addresses are together
 *
 * MAC flapping detection is hard without knowledge of the topology
 * (algorithm deadlock)
 *
 * but topology discovery can't be based solely on leaf ports
 * there are a transit only switches in the wild
 *
 * so I'll give it a try with the former algorithm, which is more general
 *
 * 2nd step: discovering neighbouring bridges
 *
 * idea: neighbours have no intersection
 *
 * - if opposite bridges are found on two ports
 *   and this ports have not common addresses
 *   then this ports are strong neighbours (connected)
 *   data: port->neighbour
 *
 * in the case there are not strong neighbours (hubs or unknown bridges exist)
 * - if opposite bridges are found on two ports
 *   and this ports have not common bridges
 *   then this ports are weak neighbours (connected via HoUB)
 *   - there may be more than one such neighbour
 *     (HoUB works as multipoint connection)
 *   data: port->neighbours
 *
 *
 * linear topologies (two-port bridges connected together as vagons in a train)
 * can not be discovered without additional information outside of FDB
 * (management MAC on independent port, LLDP...)
 *
 * it is possible to reveal orientation of such bridges towards other bridges
 * but not the right order
 */

#include "process.h"

#include "../lib/list.h"
#include "../lib/log.h"

// printf()
#include <stdio.h>

// sleep()
#include <unistd.h>

// atol()
#include <stdlib.h>

// str*()
#include <string.h>

// number of seconds to wait if malloc() failed
#define MALLOC_FAILED_DELAY	10

/*
 * data declarations (may land in data.h in the future)
 *
 * all ids must be on the first place to simplify search function
 */

// for type conversion
struct vertex {
	char *id;
};

struct bridge {
	char *id;
	struct list_element *ports;
};

/*
 * bridge - points to host bridge
 * time-addresses - list of addresses on this port marked with time
 * 	(weighted edges)
 * bridges - list of bridges behind this port (partial result of the algorithm)
 * neighbour - probable neighbour (final result of the algorithm)
 * neighbours - weak / far neighbours because of unknown bridge on this port
 */
struct port {
	char *id;
	struct bridge *bridge;
	//struct list_element *time_addresses;
	struct list_element *addresses;
	struct list_element *bridges;
	struct port *neighbour;
	struct list_element *neighbours;
};

/*
 * ports - list of all ports where this address was seen
 * leaf_port - port on which is this address unique
 */
struct address {
	char *id;
	struct list_element *ports;
	struct port *leaf_port;
};

/*
 * store the time for the address (used on the port)

struct time_address {
	long time;
	struct address *address;
};

*/

// time - port - address struct is for time saving and aging (later)
struct tpa {
	long time;
	struct port *port;
	struct address *address;
};


// data definitions

// list of bridge lists (each element is a physical network)
//static struct list_element *networks;

// all bridges in one list
static struct list_element *bridges = NULL;

// all addresses in one list
static struct list_element *addresses = NULL;

// list of time stamped port - address pairs
static struct list_element *tpas_head = NULL, *tpas_tail = NULL;


// functions definitions

void *find_string_id_in_list(const char *string, struct list_element *head)
{
	while (head) {
		// assuming that data starts with string id
		if (! strcmp(string, ((struct vertex *)head->data)->id))
			return head->data;
		head = head->next;
	}

	return NULL;
}

void print_string_ids_from_list(struct list_element *head)
{
	int counter = 0;

	while (head) {
		printf("#%d %s\n", ++counter,
				((struct vertex *)head->data)->id);
		head = head->next;
	}

	sleep(1);
}

char *copy_string(const char *src)
{
	char *dst;

	while (! (dst = strdup(src))) {
		log_write("copy_string(): strdup() failed\n");
		sleep(MALLOC_FAILED_DELAY);
	}

	return dst;
}

// encapsulate list_insert() error handling
void insert_into_list(void *data, struct list_element **head)
{
	while (list_insert(head, data)) {
		log_write("insert_into_list(): list_insert() failed\n");
		sleep(MALLOC_FAILED_DELAY);
	}
}

void *create_vertex(size_t size, const char *id, struct list_element **head)
{
	struct vertex *vertex;

	// create a new vertex with appropriate size
	while (! (vertex = malloc(size))) {
		log_write("create_vertex(): malloc() failed\n");
		sleep(MALLOC_FAILED_DELAY);
	}

	// insert string id in the new vertex
	vertex->id = copy_string(id);

	// insert the new vertex in desired list
	insert_into_list(vertex, head);

	return vertex;
}

// reverse function to create_vertex()
void delete_vertex(struct vertex *vertex, struct list_element **head)
{
	// delete the vertex from given list
	list_delete_data(head, vertex);

	// release string id (allocated by strdup() in copy_string())
	free(vertex->id);

	// release the vertex
	free(vertex);
}

void print_address(struct address *address)
{
	printf("\t\t%s\n", address->id);
}

void print_bridges(struct list_element *bridges_to_print)
{
	struct list_element *this = bridges_to_print;

	while (this) {
		printf("\t\tbridge %s\n", ((struct bridge *)this->data)->id);
		this = this->next;
	}
}

// print port - bridges, neighbour, neighbours
void print_port(struct port *port)
{
	struct list_element *this;
	struct bridge *bridge;
	struct port *nport;

	// do not print leaf ports
	if (! port->bridges)
		return;

	printf("bridge %s, port %s", port->bridge->id, port->id);

	if (port->neighbour)
		printf(" - port %s, bridge %s",
				port->neighbour->id,
				port->neighbour->bridge->id);

	printf("\nbridges:");
	this = port->bridges;
	while (this) {
		bridge = this->data;
		printf(" %s", bridge->id);
		this = this->next;
	}

	if (port->neighbours)
		printf("\nneighbours:");
	this = port->neighbours;
	while (this) {
		nport = this->data;
		printf(" %s@%s", nport->id, nport->bridge->id);
		this = this->next;
	}

	printf("\n");
}

void print_bridge(struct bridge *bridge)
{
	struct list_element *this;

	printf("%s\n", bridge->id);

	this = bridge->ports;
	while (this) {
		print_port(this->data);
		this = this->next;
	}
}

void print_add_neighbours_into_db(struct port *port_a, struct port *port_b)
{
	printf("insert fdb2graph_neighbour (agent_a, port_a, port_b, agent_b)"
			" values (%s, %s, %s, %s);\n",
			port_a->bridge->id, port_a->id,
			port_b->id, port_b->bridge->id);
	printf("insert fdb2graph_neighbour (agent_a, port_a, port_b, agent_b)"
			" values (%s, %s, %s, %s);\n",
			port_b->bridge->id, port_b->id,
			port_a->id, port_a->bridge->id);
}

void print_del_neighbours_from_db(struct port *port_a, struct port *port_b)
{
	printf("delete from fdb2graph_neighbour where agent_a = %s"
	" and port_a = %s and port_b = %s and agent_b = %s;\n",
			port_a->bridge->id, port_a->id,
			port_b->id, port_b->bridge->id);
	printf("delete from fdb2graph_neighbour where agent_a = %s"
	" and port_a = %s and port_b = %s and agent_b = %s;\n",
			port_b->bridge->id, port_b->id,
			port_a->id, port_a->bridge->id);
}

void print_add_leafport_into_db(struct address *address)
{
	printf("insert fdb2graph_leafport (mac, agent, port)"
			" values ('%s', %s, %s)"
			" on duplicate key update agent = %s, port = %s;\n",
			address->id,
			address->leaf_port->bridge->id,
			address->leaf_port->id,
			address->leaf_port->bridge->id,
			address->leaf_port->id);
}

void print_del_leafport_from_db(struct address *address)
{
	printf("delete from fdb2graph_leafport where"
			" mac = '%s' and agent = %s and port = %s;\n",
			address->id,
			address->leaf_port->bridge->id,
			address->leaf_port->id);
}

void disconnect_ports(struct port *port_a, struct port *port_b)
{
	if (port_a->neighbour != port_b || port_b->neighbour != port_a) {
		// port_a is not connected with port_b
		printf("not connected: bridge %s - %s - %s - bridge %s\n",
				port_a->bridge->id, port_a->id,
				port_b->id, port_b->bridge->id);
		return;
	}

	// disconnect port_a and port_b
	port_a->neighbour = port_b->neighbour = NULL;
	printf("disconnected: bridge %s - %s x %s - bridge %s\n",
			port_a->bridge->id, port_a->id,
			port_b->id, port_b->bridge->id);

	// output for MySQL
	print_del_neighbours_from_db(port_a, port_b);
}

void neighbours_connect(struct port *port_a, struct port *port_b)
{
	if (list_contains(port_a->neighbours, port_b)
		&& list_contains(port_b->neighbours, port_a)) {

		//printf("ports already connected\n");
		return;
	}

	// connect
	insert_into_list(port_b, &port_a->neighbours);
	insert_into_list(port_a, &port_b->neighbours);

	// output for MySQL
	print_add_neighbours_into_db(port_a, port_b);
}

void neighbours_disconnect(struct port *port_a, struct port *port_b)
{

	// may be useless check
	if (! list_contains(port_a->neighbours, port_b)
		|| ! list_contains(port_b->neighbours, port_a)) {

		printf("ports not connected\n");
		return;
	}

	// disconnect
	list_delete_data(&port_a->neighbours, port_b);
	list_delete_data(&port_b->neighbours, port_a);

	// output for MySQL
	print_del_neighbours_from_db(port_a, port_b);
}

void neighbours_disconnect_all(struct port *port_a)
{
	struct list_element *this = port_a->neighbours;
	struct port *port_b;

	while (this) {

		port_b = this->data;

		// use this now, it will be free after neighbours_disconnect()
		this = this->next;

		neighbours_disconnect(port_a, port_b);
	}
}

void connect_ports(struct port *port_a, struct port *port_b)
{
	if (port_a->neighbour == port_b && port_b->neighbour == port_a) {

		/*
		printf("already connected: bridge %s - %s - %s - bridge %s\n",
				port_a->bridge->id, port_a->id,
				port_b->id, port_b->bridge->id);
		*/
		return;
	}

	// if port_a is connected to another port than port_b, disconnect it
	if (port_a->neighbour) disconnect_ports(port_a, port_a->neighbour);

	// the same for port_b
	if (port_b->neighbour) disconnect_ports(port_b, port_b->neighbour);

	// if port_a has a weak neghbours, clean them all
	if (port_a->neighbours) neighbours_disconnect_all(port_a);

	// the same for port_b
	if (port_b->neighbours) neighbours_disconnect_all(port_b);

	// connect port_a and port_b
	port_a->neighbour = port_b;
	port_b->neighbour = port_a;
	printf("connected: bridge %s - %s - %s - bridge %s\n",
			port_a->bridge->id, port_a->id,
			port_b->id, port_b->bridge->id);

	// output for MySQL
	print_add_neighbours_into_db(port_a, port_b);
}

void delete_address(struct address *address)
{
	printf("deleting address %s\n", address->id);
	delete_vertex((void *)address, &addresses);
}

struct port *find_bridge_on_bridge(struct bridge *whatb, struct bridge *whereb)
{
	struct list_element *this_port = whereb->ports, *this_bridge;

	while (this_port) {

		this_bridge = ((struct port *)this_port->data)->bridges;

		while (this_bridge) {

			// check if bridge we are looking for is on this_port
			if (this_bridge->data == whatb) {
				/*
				printf("bridge %s found on port %s"
						" of bridge %s\n", whatb->id,
						((struct port *)this_port->data)->id,
						((struct port *)this_port->data)->bridge->id);
				*/
				return this_port->data;
			}

			this_bridge = this_bridge->next;
		}

		this_port = this_port->next;
	}

	return NULL;
}

struct port *find_bridge_in_neighbours(struct bridge *bridge,
		struct port *port_a)
{
	struct list_element *this = port_a->neighbours;
	struct port *port_b;

	while (this) {

		port_b = this->data;

		if (port_b->bridge == bridge) return port_b;

		this = this->next;
	}

	return NULL;
}

void delete_bridge_from_port(struct bridge *bridge, struct port *port)
{
	struct port *port_b;

	// if the bridge is neighbouring via the port disconnect first
	if (port->neighbour && port->neighbour->bridge == bridge)
		disconnect_ports(port, port->neighbour);
	else
		if ((port_b = find_bridge_in_neighbours(bridge, port))) {

			// the bridge is in neighbours, disconnect
			neighbours_disconnect(port, port_b);
		}

	printf("deleting bridge %s from port %s on bridge %s\n",
			bridge->id, port->id, port->bridge->id);

	// delete the bridge from the port
	list_delete_data(&port->bridges, bridge);
}

// clean up the bridge from port->bridges on each bridge
void remove_bridge_from_all_port_bridges(struct bridge *bridge)
{
	struct list_element *this = bridges;
	struct port *port;

	while (this) {
		port = find_bridge_on_bridge(bridge, this->data);
		if (port) delete_bridge_from_port(bridge, port);
		this = this->next;
	}
}

void delete_bridge(struct bridge *bridge)
{
	remove_bridge_from_all_port_bridges(bridge);

	// remove bridge from list of all bridges
	list_delete_data(&bridges, bridge);

	printf("deleting bridge %s\n", bridge->id);
	delete_vertex((void *)bridge, &bridges);
}

void delete_port(struct port *port)
{
	struct bridge *bridge = port->bridge;

	// cleanup port->neighbour (break neighbourhood)
	if (port->neighbour) disconnect_ports(port, port->neighbour);

	// cleanup port->neighbours (break possible other neighbourhoods)
	neighbours_disconnect_all(port);

	// cleanup port->bridges
	printf("deleting bridges from port %s on bridge %s\n",
			port->id, bridge->id);
	while (port->bridges)
		list_delete_head(&port->bridges);

	// cleanup port
	printf("deleting port %s from the bridge %s\n", port->id, bridge->id);
	delete_vertex((void *)port, &bridge->ports);

	// remove bridge with no ports
	if (! bridge->ports) delete_bridge(bridge);
}

void set_leaf_port(struct address *address, struct port *port)
{
	// unconditionally set the desired port as leaf_port for the address
	address->leaf_port = port;

	print_add_leafport_into_db(address);
}

void unset_leaf_port(struct address *address, struct port *port)
{
	if (address->leaf_port == port) {
		// yes, the port is set as leaf port for the address

		// first send SQL delete
		// do not delete, just rely on update and save history
		//print_del_leafport_from_db(address);

		// then unset it
		address->leaf_port = NULL;
	}
}

void update_leaf_port_after_del(struct address *address, struct port *port)
{
	if (! port->addresses) {
		// last one address was removed from the port
		unset_leaf_port(address, port);
	} else
		if (! port->addresses->next) {
			// one address on port
			printf ("port %s on bridge %s became leaf port\n",
					port->id, port->bridge->id);
			/*
			 * the recent port with one address
			 * became the leaf port for the address
			 */
			set_leaf_port(port->addresses->data, port);
		}
}

// remove only, do not call delete_address() and delete_port (free())
void remove_address_from_port(struct address *address, struct port *port)
{
	// delete the port from the address
	list_delete_data(&address->ports, port);

	// delete address from the port
	list_delete_data(&port->addresses, address);

	// update the leaf port for the last or removed address
	update_leaf_port_after_del(address, port);
}

void delete_address_from_port(struct address *address, struct port *port)
{
	remove_address_from_port(address, port);

	// delete the address, if there is no port with this address
	if (! address->ports) delete_address(address);

	// delete the port from a bridge, if the port has no address
	if (! port->addresses) delete_port(port);
}

// remote_bridge must be located behind the local_port
struct port *find_opposite_port_on_remote_bridge(
		struct port *local_port, struct bridge *remote_bridge)
{
	struct port *opposite_port, *port_a, *port_b;
	struct list_element *this_a, *this_b;

	/*
	 * first try to find bridge_a in port->bridges on bridge_b
	 * it should be faster
	 */
	opposite_port = find_bridge_on_bridge(local_port->bridge,
			remote_bridge);

	if (opposite_port)
		return opposite_port;

	// now try to find opposite port using MAC addresses

	this_a = local_port->bridge->ports;

	while (this_a) {

		port_a = this_a->data;

		// avoid local_port when inspecting addresses
		if (port_a == local_port) {

			this_a = this_a->next;
			continue;
		}

		this_b = remote_bridge->ports;

		while (this_b) {

			port_b = this_b->data;

			/*
			 * if some port_a on bridge_a other than local_port
			 * has common address(es) with some port_b on bridge_b
			 * than port_b is opposite port
			 */
			if (lists_share_data(port_a->addresses,
						port_b->addresses))
				return port_b;

			this_b = this_b->next;
		}

		this_a = this_a->next;
	}

	// opposite port was not found
	return NULL;
}

void check_if_bridge_is_neighbour_and_update_neighbourhood(
		struct port *port_a, struct bridge *bridge_b)
{
	struct port *port_b;

	//port_b = find_bridge_on_bridge(port_a->bridge, bridge_b);
	port_b = find_opposite_port_on_remote_bridge(port_a, bridge_b);

	//printf("check_if_bridge_is_neighbour_and_update_neighbourhood()\n");

	if (! port_b) {

		/*
		printf("bridge_a %s was not found on the bridge_b %s\n",
				port_a->bridge->id, bridge_b->id);
		*/

		return;
	}

	if (! lists_share_data(port_a->addresses, port_b->addresses)) {

		// there is no intersection in addresses on port_a and port_b
		connect_ports(port_a, port_b);

	} else {

		// there is another bridge between bridge_a and bridge_b

		// leave if neighbour bridge is already known
		if (port_a->neighbour) {

			/*
			printf("bridge %s port %s has already"
					" neighbour port %s bridge %s\n",
					port_a->bridge->id, port_a->id,
					port_a->neighbour->id,
					port_a->neighbour->bridge->id);
			*/

			return;
		}

		if (! lists_share_data(port_a->bridges, port_b->bridges)) {

			/*
			 * there is no intersection in bridges
			 * on port_a and port_b
			 *
			 * may be this is the time to verify
			 * other port_a->neighbours if they are still valid
			 * and remove obsolete ones
			 */
			neighbours_connect(port_a, port_b);
		}
	}
}

void update_bridge_on_opposite_port_too(struct port *port_a,
		struct bridge *bridge_b)
{
	struct port *port_b;

	// nothing to do - bridge_a is already presented on bridge_b
	if (find_bridge_on_bridge(port_a->bridge, bridge_b))
		return;

	/*
	 * this is ugly - find_bridge_on_bridge() is called again
	 * in find_opposite_port_on_remote_bridge()
	 * and find_opposite_port_on_remote_bridge() is called again
	 * in check_if_bridge_is_neighbour_and_update_neighbourhood()
	 * optimize it later, if this function improves the algorithm
	 */
	port_b = find_opposite_port_on_remote_bridge(port_a, bridge_b);

	if (port_b) {

		// opposite port found, so add bridge_a to its list
		printf("adding bridge %s to port %s on bridge %s\n",
				port_a->bridge->id, port_b->id, bridge_b->id);

		insert_into_list(port_a->bridge, &port_b->bridges);
	}
}

void add_bridge_to_port(struct bridge *bridge, struct port *port)
{
	printf("adding bridge %s to port %s on bridge %s\n",
			bridge->id, port->id, port->bridge->id);

	// add the bridge to the port
	insert_into_list(bridge, &port->bridges);

	printf("add_bridge_to_port() bridge: %s, port: %s, bridge to add: %s\n",
			port->bridge->id, port->id, bridge->id);
	//print_bridges(port->bridges);

	/*
	 * check if opposite port->bridges contain this bridge
	 * and if it doesn'n try to add this bridge to opposite port->bridges
	 */
	update_bridge_on_opposite_port_too(port, bridge);

	// the bridge added may be a new neighbour
	check_if_bridge_is_neighbour_and_update_neighbourhood(port, bridge);
}

void update_bridge_on_port(struct bridge *bridge, struct port *port)
{
	struct port *old_port = find_bridge_on_bridge(bridge, port->bridge);

	if (old_port) {
		// bridge was found on searched bridge
		if (old_port == port) {

			// this may be necessary
			check_if_bridge_is_neighbour_and_update_neighbourhood(
					port, bridge);

			// bridge already presented on the port, we are done
			return;

		} else {

			/*
			 * bridge found on another port of the searched bridge
			 * is this correct to solve uniqueness here?
			 * what about circle topologies?
			 */
			delete_bridge_from_port(bridge, old_port);
		}
	}

	add_bridge_to_port(bridge, port);
}

void check_address_pair(struct address *address_a, struct address *address_b,
		struct port *port)
{
	struct list_element *this_a = address_a->ports, *this_b;

	//printf("checking addresess: %s %s\n", address_a->id, address_b->id);

	while (this_a) {
		this_b = address_b->ports;
		while (this_b) {
			if (((struct port *)this_a->data)->bridge
					== ((struct port *)this_b->data)->bridge
					&& this_a->data != this_b->data) {

				/*
				printf("bridge %s\n",
						((struct port *)this_a->data)->bridge->id);
				printf("\tport %s address %s\n",
						((struct port *)this_a->data)->id,
						address_a->id);
				printf("\tport %s address %s\n",
						((struct port *)this_b->data)->id,
						address_b->id);
				*/

				//sleep(3);

				/*
				 * a bridge was found (this_a->data->bridge)
				 * which has the inspecting addresses
				 * on different ports
				 *
				 * update this bridge on the uplink port
				 * of the bridge
				 * which has the inspecting addresses
				 * on the same port
				 */
				update_bridge_on_port(((struct port *)this_a->data)->bridge, port);

				// common bridge found, skip to the next port
				break;
			}
			this_b = this_b->next;
		}
		this_a = this_a->next;
	}
}

void check_address_pairs(struct address *address, struct port *port)
{
	struct list_element *this = port->addresses;

	while (this) {
		check_address_pair(address, (struct address *)this->data, port);
		this = this->next;
	}
}

void update_leaf_port_before_add(struct address *address, struct port *port)
{
	if (! port->addresses) {
		// no address on port, the first will be added
		printf("port %s on bridge %s became leaf port"
				" with address %s\n",
				port->id, port->bridge->id, address->id);
		/*
		 * the recent port with one address
		 * became the leaf port for the address
		 */
		set_leaf_port(address, port);
	} else
		if (! port->addresses->next) {
			// one address on port, the second will be added
			printf ("port %s on bridge %s became transit port\n",
					port->id, port->bridge->id);
			unset_leaf_port(port->addresses->data, port);
		}
}

void add_address_to_port(struct address *address, struct port *port)
{
	/*
	 * try to find bridges on this port
	 * do it before adding the address to port to avoid useless loop
	 */

	/*
	 * turn off this hook to topology discovery
	 */
	//check_address_pairs(address, port);

	printf("add %s to port %s on bridge %s\n",
			address->id, port->id, port->bridge->id);

	// update address->leaf_port if necessary
	update_leaf_port_before_add(address, port);

	// add the address to the port
	insert_into_list(address, &port->addresses);

	// add the port to the address
	insert_into_list(port, &address->ports);
}

void add_tpa(long time, struct port *port, struct address *address)
{
	struct tpa *tpa;

	// allocate a new struct tpa
	while (! (tpa = malloc(sizeof(struct tpa)))) {
		log_write("add_tpa(): malloc() failed\n");
		sleep(MALLOC_FAILED_DELAY);
	}

	// fill the struct
	tpa->time = time;
	tpa->port = port;
	tpa->address = address;

	// append the tpa on the end of the list
	if (tpas_tail) {
		// list is not empty
		while (list_append(tpas_tail, tpa)) {
			log_write("add_tpa(): list_append() failed\n");
			sleep(MALLOC_FAILED_DELAY);
		}

		// update tpas_tail
		tpas_tail = tpas_tail->next;

	} else {
		// list is empty
		while (list_insert(&tpas_head, tpa)) {
			log_write("add_tpa(): list_insert() failed\n");
			sleep(MALLOC_FAILED_DELAY);
		}

		// update tpas_tail
		tpas_tail = tpas_head;
	}
}

void delete_all_pa_from_tpas(struct port *port, struct address *address)
{
	struct list_element *prev, *this = tpas_head;
	struct tpa *tpa;

	while (this) {
		tpa = this->data;

		if (tpa->port == port && tpa->address == address) {

			// delete this list_element

			// delete this->data first
			free(tpa);

			// delete this
			if (this == tpas_head) {

				// still at the beginning
				list_delete_head(&tpas_head);

				if (tpas_head) {
					// the list is not empty
					// set this to the new head
					this = tpas_head;
					continue;
				} else {
					// the list is empty
					// set tail properly
					tpas_tail = NULL;
					return;
				}
			} else {

				// deleting the second or above list_element
				list_delete_after(prev);

				if (prev->next) {
					// there is another element in the list
					this = prev->next;
					continue;
				} else {
					// last element was deleted
					tpas_tail = prev;
					return;
				}
			}
		}

		prev = this;
		this = this->next;
	}
}

/*
 * check if the address is already on this bridge
 * and update if the address is on another port
 * use list of ports where address is presented (address->ports)
 */
void update_address_on_bridge(struct address *address, struct port *new_port)
{
	struct list_element *this = address->ports;
	struct port *port;

	while (this) {

		port = this->data;

		if (port->bridge == new_port->bridge) {

			// the address is on this bridge

			/*
			 * the address is already on this port
			 * nothing to update - leave this function
			 */
			if (port == new_port) return;

			// the address is on another port of the bridge

			printf("address %s moved from %s to %s"
					" on the bridge %s\n",
					address->id, port->id, new_port->id,
					port->bridge->id);

			/*
			 * remove this pa binding from all tpas
			 * to save consistency
			 * and avoid use after free in aging()
			 */
			delete_all_pa_from_tpas(port, address);

			// remove it from old port (preserve uniqueness)
			remove_address_from_port(address, port);

			/*
			 * delete the port from a bridge,
			 * if the port has no address left
			 * only this one delete (in process_row() chain)
			 * is leading to free(); all others are in aging()
			 */
			if (! port->addresses) delete_port(port);

			/*
			 * the address is always unique on one bridge
			 * and it's found right now, so leave this loop
			 */
			break;
		}

		this = this->next;
	}

	// add the address to the current port
	add_address_to_port(address, new_port);
}

void process_row(const char *bridge_id, const char *port_id,
		const char *address_id,	long time)
{
	struct bridge *bridge;
	struct port *port;
	struct address *address;

	//printf("%s %s %s %ld\n", bridge_id, port_id, address_id, time);

	if ((bridge = find_string_id_in_list(bridge_id, bridges))) {
		//printf("bridge %s found in bridges list\n", bridge_id);
	} else {
		printf("bridge %s not found in bridges list\n", bridge_id);
		//print_string_ids_from_list(bridges);
		bridge = create_vertex(sizeof(struct bridge), bridge_id,
				&bridges);

		// init
		bridge->ports = NULL;
	}

	if ((port = find_string_id_in_list(port_id, bridge->ports))) {
		//printf("port %s found at bridge %s\n", port_id, bridge->id);
	} else {
		printf("port %s not found at bridge %s\n", port_id, bridge->id);

		port = create_vertex(sizeof(struct port), port_id,
				&bridge->ports);

		// init
		port->bridge = bridge;
		port->addresses = NULL;
		port->bridges = NULL;
		port->neighbour = NULL;
		port->neighbours = NULL;
	}

	if ((address = find_string_id_in_list(address_id, addresses))) {
		//printf("address %s found in addresses list\n", address_id);
	} else {
		printf("address %s not found in addresses list\n", address_id);

		address = create_vertex(sizeof(struct address), address_id,
		&addresses);

		// init
		address->ports = NULL;
		address->leaf_port = NULL;
	}

	update_address_on_bridge(address, port);

	add_tpa(time, port, address);

	//print_bridge(bridge);
}

void process_data(MYSQL_RES *result)
{
	int fields = mysql_num_fields(result);
	MYSQL_ROW row;

	// expecting four columns in result
	if (fields != 4) return;

	puts("\n");

	while ((row = mysql_fetch_row(result))) {
		process_row(row[0], row[1], row[2], atol(row[3]));
		//sleep(1);
	}

	printf("count of the bridges: %ld\n", list_count(bridges));
	printf("count of the addresses: %ld\n", list_count(addresses));
	printf("count of the tpas: %ld ", list_count(tpas_head));
	if (tpas_head)
		printf("the oldest entry: %ld",
				((struct tpa *)tpas_head->data)->time);
	printf("\n");
}

int tpa_head_is_unique(void)
{
	struct tpa *tpa_head = tpas_head->data, *tpa;

	// start from the second element
	struct list_element *this = tpas_head->next;

	while (this) {
		tpa = this->data;

		// if not unique
		if (tpa->port == tpa_head->port
				&& tpa->address == tpa_head->address)
			return 0;

		this = this->next;
	}

	// the first element is unique
	return 1;
}

/*
 * the time is neccessary for topology changes which involve broken links
 * another approach is to use event driven SNMP data (MAC add/del)
 * but this require stateful SNMP reading; db has time, so use it
 *
 * delete and free() obsolete addresses, ports, bridges
 * as opposite to process_row()
 */
void aging(long time)
{
	struct tpa *tpa;

	while (tpas_head) {

		// just to simplify references
		tpa = tpas_head->data;

		// no more of old data, so end this job
		if (tpa->time > time) return;

		if (tpa_head_is_unique())
			delete_address_from_port(tpa->address, tpa->port);

		// release the old tpa struct
		free(tpa);

		// delete the head
		list_delete_head(&tpas_head);

	}

	// list is empty, update the tail pointer
	tpas_tail = NULL;
}

// for all combinations of two addresses from port->addresses
void topology_port(struct port *port)
{
	struct list_element *this_a = port->addresses, *this_b;
	struct address *address_a, *address_b;

	// for all addresses in port->addresses list
	while (this_a) {

		address_a = this_a->data;

		// take the next address
		this_b = this_a->next;

		// and for this and all next
		while (this_b) {

			address_b = this_b->data;

			// call the former algorithm above
			check_address_pair(address_a, address_b, port);

			this_b = this_b->next;
		}

		this_a = this_a->next;
	}

	print_port(port);
}

void topology_bridge(struct bridge *bridge)
{
	struct list_element *this = bridge->ports;

	while (this) {

		topology_port(this->data);

		this = this->next;
	}
}

void topology_bridges(void)
{
	struct list_element *this = bridges;

	while (this) {

		topology_bridge(this->data);

		this = this->next;
	}
}

/*
 * a topology behind a port can change without an address add/delete
 * on this port; then topology discovery can't be triggered by process_data()
 * so here we are with independent process_topology()
 *
 * again - topology discovery trigered by address add doesn't work
 * if addresses are stable on the bridge
 * because the second bridge can know the first bridge
 * (rows from the first bridge are already processed),
 * but the first bridge doesn't know about the second bridge
 * until an address is added to the first bridge
 * after the second bridge is processed
 * - so process_topology() is a must in this phase
 */
void process_topology(void)
{
	topology_bridges();
}
