#include "common.h"
#include "list.h"
#include "capture.h"
#include "packet.h"
#include "output.h"
#include "confile.h"
#include "log.h"
#include "db.h"
#include "queue.h"

// fprintf
#include <stdio.h>
#include <signal.h>
// daemon()
#include <unistd.h>
// malloc(), free()
#include <stdlib.h>
// strcpy(), strcat()
#include <string.h>

// pcap_findalldevs is a quite expensive
// so it will be called once per IFACE_SCAN_PERIOD * packet_loop()
#define IFACE_SCAN_PERIOD	60
// some limit when database is gone 
#define DESIRED_LIST_SIZE	10000
#define MAXIMUM_LIST_SIZE	11000

// visible here, non optimized
static volatile sig_atomic_t lastsig = 0;

char *append_conf_suffix(char *basename)
{
	char *confname, suffix[] = ".conf";
	confname = malloc(strlen(basename) + strlen(suffix) + 1);
	if (confname != NULL) {
		strcpy(confname, basename);
		strcat(confname, suffix);
	}
	return confname;
}

void print_usage(void)
{
	fputs("ND and ARP Listener for nana Network Analyser\n", stderr);
	fputs("Usage: nals [config-file]\n", stderr);
}

static void signal_handler(int signo)
{
	lastsig = signo;
}

// formerly named shutdown - it leads to sigsegv, when mysql_real_connect
// calls this function instead of libc one; I don't understand linker
// behavior and weak symbols so rename was the simpliest solution
void close_log_and_release_conf(struct conf *conf) {
	log_print_time();
	fputs("exiting\n", logfile);
	log_close_logfile();
	conf_release(conf);
}

int main (int argc, char **argv)
{
	char *confname;
	struct conf *conf;
	int can_connect = 0, connected = 0, i;
	struct capture *captures = NULL;
	struct queue *queue;
	switch (argc) {
		case 1:
			confname = append_conf_suffix(argv[0]);	
			if (confname == NULL) {
				fputs("malloc() failed\n", stderr);
				return 1;
			} else {
				conf = conf_get(confname);
				free(confname);
			}
			break;
		case 2:
			conf = conf_get(argv[1]);
			break;
		default:
			print_usage();
			return 1;
	}
	if (conf == NULL)
		logfile = stderr;
	else {
		log_open_logfile(conf->logfilename);
		if ((conf->dbhost != NULL)
				&& (conf->dbname != NULL)
				&& (conf->dbuser != NULL)
				&& (conf->dbpass != NULL))
			can_connect = 1;
	}
	// queue_init() use logfile, so here we go
	queue = queue_init();
	if (queue == NULL) {
		close_log_and_release_conf(conf);
		return 1;
	}
	if (! can_connect) {
		log_print_time();
		fputs("missing dbhost, dbname, dbuser or dbpass\n", logfile);
		log_print_time();
		fputs("please check config-file\n", logfile);
		log_print_time();
		fputs("printing queries to stdout\n", logfile);
	}
	if ((logfile != stderr) && can_connect) {
		log_print_time();
		fputs("calling daemon()\n", logfile);
		if (daemon(0, 0))
			fputs("daemon() failed\n", logfile);
	} else {
		log_print_time();
		fputs("can not daemonize\n", logfile);
	}
	if (signal(SIGTERM, signal_handler) == SIG_ERR) {
		log_print_time();
		fputs("fail to bind to SIGTERM via signal()\n", logfile);
	}
	while (lastsig != SIGTERM) {
		upkeep_captures(&captures);
		if (! connected && can_connect) {
			log_print_time();
			fprintf(logfile, "not connected, trying\n");
			connected = db_connect(conf);
			//fprintf(logfile, "connected: %d\n", connected);
		}
		for (i = 0; i < IFACE_SCAN_PERIOD; i++) {
			if (lastsig == SIGTERM) break;
			packet_loop(captures, queue);
			if (lastsig == SIGTERM) break;
			if (can_connect) {
				if (connected)
					connected = db_flush_queue(queue);
				else
					queue_shrink(queue, MAXIMUM_LIST_SIZE,
							DESIRED_LIST_SIZE);
			} else
				print_queue(queue);
			fflush(logfile);
		}
	}
	db_close();
	free_captures(captures);
	queue_destroy(queue);
	close_log_and_release_conf(conf);
	return 0;
}
